# ATTENTION:
# This file is generated by exasol/toolbox/pre_commit_hooks/package_version.py when using:
#   * either "poetry run nox -s fix"
#   * or "poetry run version-check <path/version.py> --fix"
# Do not edit this file manually!
# If you need to change the version, do so in the project.toml, e.g. by using `poetry version X.Y.Z`.
MAJOR = 0
MINOR = 6
PATCH = 1
VERSION = f"{MAJOR}.{MINOR}.{PATCH}"
